from typing import Union


def dprint(*args, **kwargs):
    import traceback

    """
    Pre-pends the filename and linenumber to the print statement
    """
    try:
        stack = traceback.extract_stack()[:-1]
        last = stack[-1]

        # Handle different versions of the traceback module
        if hasattr(last, "filename"):
            _filename = last.filename.split("lib")
            filename = "lib".join(_filename[1:])[1:] if len(_filename) >= 2 else last.filename
            out_str = f"\t\033[90m{filename}:{last.lineno}\033[0m"
        else:
            _filename = last[0].split("lib")
            filename = "lib".join(_filename[1:])[1:] if len(_filename) >= 2 else last[0]
            out_str = f"\t\033[90m{filename}:{last[1]}\033[0m"

        # Prepend the filename and linenumber
        __builtins__["oldprint"](*args, out_str, **kwargs)
    except:
        __builtins__["oldprint"](*args, **kwargs)


def override():
    if "oldprint" not in __builtins__:
        __builtins__["oldprint"] = __builtins__["print"]

    __builtins__["print"] = dprint


def regpid():
    import os

    id = 0

    tmpdir = os.environ.get("TMPDIR")
    if tmpdir is None:
        tmpdir = os.environ.get("TEMP")
    if tmpdir is None:
        tmpdir = os.environ.get("TMP")
    if tmpdir is None:
        tmpdir = "."

    vtarget_pids_path = os.path.join(tmpdir, "vtarget-pids")

    print(vtarget_pids_path)

    with open(vtarget_pids_path, "a+") as f:
        id = os.getpid()
        f.write(f"{id}\n")
    return id


def syspath():
    import sys

    if getattr(sys, "frozen", False):
        import os

        current_path = os.path.dirname(sys.executable)
        sys.path.insert(0, os.path.join(current_path, "python"))
        sys.path.insert(0, os.path.join(current_path, "python", "lib"))
        sys.path.insert(0, os.path.join(current_path, "python", "lib", "site-packages"))
        sys.path.insert(0, os.path.join(current_path, "python", "lib", "python3", "site-packages"))
        sys.path.insert(
            0, os.path.join(current_path, "python", "lib", "python3.10", "site-packages")
        )
        sys.path.insert(
            0, os.path.join(current_path, "python", "lib", "python3.11", "site-packages")
        )
        sys.path.insert(
            0, os.path.join(current_path, "python", "lib", "python3.12", "site-packages")
        )
        sys.path.insert(0, os.path.join(current_path, "python", "DLLs"))


def get_serial_number() -> Union[bytes, None]:
    import platform
    import subprocess
    import uuid

    os_type = platform.system()
    try:
        if os_type == "Windows":
            result = subprocess.check_output("wmic bios get serialnumber", shell=True)
            result = result.decode("utf-8").strip().split("\n")[1]
            return result
        elif os_type == "Linux":
            result = subprocess.check_output("dmidecode -s system-serial-number", shell=True)
            result = result.decode("utf-8").strip()
            return result
        elif os_type == "Darwin":
            result = subprocess.check_output("system_profiler SPHardwareDataType", shell=True)
            result = result.decode("utf-8").strip()
            lines = result.split("\n")
            serial_number_line = [line for line in lines if "Serial Number" in line][0]
            serial_number = serial_number_line.split(":")[-1].strip()
            return serial_number
        else:
            raise Exception(os_type)
    except:
        mac_address = hex(uuid.getnode()).replace("0x", "").upper()
        mac_address = ":".join(mac_address[i : i + 2] for i in range(0, 11, 2))
        return f"Dirección MAC en uso: {mac_address}"



import json

import pandas as pd

# from evalml.automl.automl_search import AutoMLSearch
from termcolor import colored

from vtarget.handlers.event_handler import event_handler


def find(arr, prop, value):
    for x in arr:
        if x[prop] == value:
            return x


class AutoTS_Cache:
    def __init__(self):
        self.df: pd.DataFrame = None
        self.basic_config: dict = None
        self.ts_quality = None
        self.series: list = []
        self.current_model_name: str = None
        self.current_serie_name: str = None
        self.autots_search = None

    def reset(self, key: str):
        pass

    def get_df(self):
        return self.df

    def get_series(self):
        return self.series

    def set_series(self, series: list):
        self.series = series

    def prepend_serie(self, serie: dict):
        self.series.insert(0, serie)

    def append_serie(self, serie: dict):
        self.series.append(serie)

    def prepend_model(self, model: dict, serie_name: str = None):
        if serie_name is None:
            serie_name = self.get_current_serie_name()
        serie = self.find_serie(serie_name)
        if serie is not None:
            list(serie.models).insert(0, model)

    def append_model(self, model: dict, serie_name: str = None):
        if serie_name is None:
            serie_name = self.get_current_serie_name()
        serie = self.find_serie(serie_name)
        if serie is not None:
            list(serie.models).append(model)

    def get_current_serie_name(self):
        return self.current_serie_name

    def set_current_serie_name(self, name: str):
        self.current_serie_name = name

    def is_a_serie_model(self, model_name: str, serie_name: str = None):
        if serie_name is None:
            serie_name = self.get_current_serie_name()
        model = self.find_serie_model(serie_name, model_name)
        return model is not None

    def find_serie(self, serie_name: str) -> dict:
        return find(self.series, "name", serie_name)

    def find_serie_model(self, serie_name: str, model_name: str) -> dict:
        serie_name = find(self.series, "name", serie_name)
        if (
            serie_name is not None
            and serie_name["models"] is not None
            and len(serie_name["models"]) > 0
        ):
            return find(serie_name["models"], "name", model_name)

    def get_current_model_name(self):
        return self.current_model_name

    def set_current_model_name(self, name: str):
        self.current_model_name = name

    def set_serie_model_prop(
        self, prop: str, value: any, serie_name: str = None, model_name: str = None
    ):
        if serie_name is None:
            serie_name = self.get_current_serie_name()
        if model_name is None:
            model_name = self.get_current_model_name()

        model = self.find_serie_model(serie_name, model_name)
        if model is not None:
            model[prop] = value
            # if prop == "status" and value == "IN_PROGRESS":
            #     self.set_current_model_name(model_name)
            event_handler.emit_queue.put(
                {
                    "name": "autots.update_model",
                    "data": json.dumps({"serie": serie_name, "model": model}, default=str),
                }
            )

    def update_serie_model(self, props: dict = {}, serie_name: str = None, model_name: str = None):
        if serie_name is None:
            serie_name = self.get_current_serie_name()
        if model_name is None:
            model_name = self.get_current_model_name()
        model = self.find_serie_model(serie_name, model_name)
        if model is not None:
            model.update(props)
            event_handler.emit_queue.put(
                {
                    "name": "autots.update_model",
                    "data": json.dumps({"serie": serie_name, "model": model}, default=str),
                }
            )

    def update_serie(self, props: dict = {}, serie_name: str = None, send_to_view: bool = True):
        if serie_name is None:
            serie_name = self.get_current_serie_name()
        serie = self.find_serie(serie_name)
        if serie is not None:
            serie.update(props)
            if send_to_view:
                event_handler.emit_queue.put(
                    {
                        "name": "autots.update_serie",
                        "data": json.dumps(
                            {k: v for k, v in serie.items() if k not in ["partition_data"]},
                            default=str,
                        ),
                    }
                )
        else:
            print(colored("NO EXISTE SERIE", "red", "on_white"))

    def get_serie_by_name(self, serie_name: str = None):
        if serie_name is None:
            serie_name = self.get_current_serie_name()
        serie = self.find_serie(serie_name)
        return serie

    def get_serie_prop(self, prop: str, serie_name: str = None):
        if serie_name is None:
            serie_name = self.get_current_serie_name()
        serie = self.find_serie(serie_name)
        if serie is not None:
            if prop in serie:
                return serie[prop]
        return None

    def set_automl_search(self, automl):
        self.autots_search = automl

    def get_autots_search(self):
        return self.autots_search


autots_cache = AutoTS_Cache()



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class CrossJoin:
    def exec(self, flow_id, node_key, pin, settings):
        script = []

        # if node_key in cache_handler.cache[flow_id] and cache_handler.cache[flow_id][node_key]['config'] == json.dumps(settings, sort_keys=True):
        # 	bug_handler.console(f'Nodo "{node_key}" leído desde cache flow_id: "{flow_id}"', 'info', flow_id)
        # 	reset_childs = False
        # 	script_handler.script += cache_handler.cache[flow_id][node_key]['script']
        # 	return cache_handler.cache[flow_id][node_key]['pout'], reset_childs

        script.append("\n# CROSS_JOIN")
        df_T: pd.DataFrame = pin["Tgt"].copy()
        df_S: pd.DataFrame = pin["Src"].copy()

        # Me quedo con los campos seleccionados del Target
        # selected_T = dict(filter(lambda x: x[1]['selected'], dtypes['Tgt']['dtypes'].items()))
        selected_T: list = settings["tgt"] if "tgt" in settings else []
        if not selected_T:
            msg = "(cross_join) Debes mantener al menos un campo en la entrada Target"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        df_T = df_T[selected_T]
        script.append("df_T = df_T[{}]".format(selected_T))

        # Me quedo con los campos seleccionados del Source
        # selected_S = dict(filter(lambda x: x[1]['selected'], dtypes['Src']['dtypes'].items()))
        selected_S: list = settings["src"] if "src" in settings else []
        if not selected_S:
            msg = "(cross_join) Debes mantener al menos un campo en la entrada Source"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        df_S = df_S[selected_S]
        script.append("df_S = df_S[{}]".format(selected_S))

        try:
            df = pd.merge(df_T, df_S, how="cross")  # , validate="many_to_one")
            script.append("df = pd.merge(df_T, df_S, how='cross')")
        except Exception as e:
            msg = "(cross_join) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import pandas as pd
from termcolor import colored

from vtarget.handlers.log_handler import log_handler


class Bug_Handler:
    def __init__(self):
        self.bug = []
        self.__append_handler = None

    def on_append(self, handler):
        self.__append_handler = handler

    def append_handler(self, bug):
        if self.__append_handler is not None:
            self.__append_handler(bug)

    def append(self, bug, use_handler=True):
        self.bug.append(bug)
        if use_handler:
            # await socket_server.emit('dataprep.bug', bug) # TODO: Mover a py
            self.append_handler(bug)

    def console(self, msg, level, flow_id, emit=True):
        level = level.upper()
        if level == "DEBUG":
            color = "green"
        elif level == "INFO":
            color = "cyan"
        elif level == "WARN":
            color = "yellow"
        elif level == "ERROR":
            color = "red"
        elif level == "FATAL":
            color = "white"
            print(colored(f"[{level}] {msg}", color, "on_red"))
            return
        elif level == "TRACE":
            color = "magenta"
        else:
            color = "white"
        print(colored(f"[{level}] {msg}", color))
        log_handler.append({"flow_id": flow_id, "level": level, "msg": msg, "color": color}, emit)

    def default_on_error(
        self,
        flow_id: str,
        node_key: str,
        msg: str,
        exception="",
        console_level="fatal",
        bug_level="error",
        success=False,
    ):
        self.console(msg, console_level, flow_id)
        self.append(
            {
                "flow_id": flow_id,
                "success": success,
                "node_key": node_key,
                "level": bug_level,
                "msg": msg,
                "exception": exception,
            }
        )
        return {"Out": pd.DataFrame()}


bug_handler = Bug_Handler()



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class DataCleansing:
    def exec(self, flow_id, node_key, pin, settings):
        script = []

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# Data Cleansing")

        # columns, replace_blanks, replace_zeros, remove_leading_trailing, remove_tabs_line_breaks, remove_all_whitespaces, remove_letters, remove_numbers, remove_punctuation, modify_case_type
        selected_columns: list = (
            settings["columns"] if ("columns" in settings and settings["columns"] is not []) else []
        )
        replace_blanks: bool = (
            settings["replace_blanks"]
            if ("replace_blanks" in settings and settings["replace_blanks"] is not None)
            else False
        )
        replace_zeros: bool = (
            settings["replace_zeros"]
            if ("replace_zeros" in settings and settings["replace_zeros"] is not None)
            else False
        )
        remove_leading_trailing: bool = (
            settings["remove_leading_trailing"]
            if (
                "remove_leading_trailing" in settings
                and settings["remove_leading_trailing"] is not None
            )
            else False
        )
        remove_tabs_line_breaks: bool = (
            settings["remove_tabs_line_breaks"]
            if (
                "remove_tabs_line_breaks" in settings
                and settings["remove_tabs_line_breaks"] is not None
            )
            else False
        )
        remove_all_whitespaces: bool = (
            settings["remove_all_whitespaces"]
            if (
                "remove_all_whitespaces" in settings
                and settings["remove_all_whitespaces"] is not None
            )
            else False
        )
        remove_letters: bool = (
            settings["remove_letters"]
            if ("remove_letters" in settings and settings["remove_letters"] is not None)
            else False
        )
        remove_numbers: bool = (
            settings["remove_numbers"]
            if ("remove_numbers" in settings and settings["remove_numbers"] is not None)
            else False
        )
        remove_punctuation: bool = (
            settings["remove_punctuation"]
            if ("remove_punctuation" in settings and settings["remove_punctuation"] is not None)
            else False
        )
        remove_nulls: bool = (
            settings["remove_nulls"]
            if ("remove_nulls" in settings and settings["remove_nulls"] is not None)
            else False
        )
        modify_case_type: str = (
            settings["modify_case_type"]
            if ("modify_case_type" in settings and settings["modify_case_type"] is not None)
            else None
        )

        try:
            columns = selected_columns if selected_columns else df.columns
            for column in columns:
                if replace_blanks and pd.api.types.is_string_dtype(df[column]):
                    df[column] = df[column].fillna("")
                    script.append(f'df["{column}"] = df["{column}"].fillna("")')
                if replace_zeros and pd.api.types.is_numeric_dtype(df[column]):
                    df[column] = df[column].fillna(0)
                    script.append(f'df["{column}"] = df["{column}"].fillna(0)')
                if remove_leading_trailing and pd.api.types.is_string_dtype(df[column]):
                    df[column] = df[column].str.strip()
                    script.append(f'df["{column}"].str.strip()')
                if remove_tabs_line_breaks and pd.api.types.is_string_dtype(df[column]):
                    df[column] = df[column].str.replace(r"[ \r\t\n]+", " ")
                    script.append(
                        f'df["{column}"] = df["{column}"].str.replace(r"[ \\r\\t\\n]+", " ")'
                    )
                if remove_all_whitespaces and pd.api.types.is_string_dtype(df[column]):
                    df[column] = df[column].str.replace(" ", "")
                    script.append(f'df["{column}"] = df["{column}"].str.replace(" ", "")')
                if remove_letters and pd.api.types.is_string_dtype(df[column]):
                    df[column] = df[column].str.replace(r"[a-zA-Z]", "")
                    script.append(f'df["{column}"] = df["{column}"].str.replace(r"[a-zA-Z]", "")')
                if remove_numbers and pd.api.types.is_string_dtype(df[column]):
                    df[column] = df[column].str.replace(r"\d", "")
                    script.append(f'df["{column}"] = df["{column}"].str.replace(r"\d", "")')
                if remove_punctuation and pd.api.types.is_string_dtype(df[column]):
                    df[column] = df[column].str.replace(r"[^\w\s]", "")
                    script.append(f'df["{column}"] = df["{column}"].str.replace(r"[^\w\s]", "")')
                if remove_nulls:
                    df = df.dropna(subset=[column])
                    script.append(f'df = df.dropna(subset=["{column}"])')

                if modify_case_type and pd.api.types.is_string_dtype(df[column]):
                    if modify_case_type == "upper_case":
                        df[column] = df[column].str.upper()
                        script.append(f'df["{column}"] = df["{column}"].str.upper()')
                    if modify_case_type == "lower_case":
                        df[column] = df[column].str.lower()
                        script.append(f'df["{column}"] = df["{column}"].str.lower()')
                    if modify_case_type == "title_case":
                        df[column] = df[column].str.title()
                        script.append(f'df["{column}"] = df["{column}"].str.title()')

        except Exception as e:
            msg = "(data_cleansing) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import json

import numpy as np
import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Switch:
    def __init__(self):
        self.functionApply = ["is null", "is not null"]
        self.noValueRequired = ["is empty", "is not empty"]

    def exec(self, flow_id, node_key, pin, settings):
        script = []

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# SWITCH")

        # cases, default_value, new_column
        cases: list = settings["cases"] if ("cases" in settings and settings["cases"] != []) else []
        default_value: str = (
            settings["default_value"]
            if ("default_value" in settings and settings["default_value"] != None)
            else None
        )
        new_column: str = (
            settings["new_column"]
            if (
                "new_column" in settings
                and settings["new_column"] != None
                and settings["new_column"] != ""
            )
            else "new_column"
        )

        if not default_value:
            msg = "(switch) Debes escribir el valor por defecto para aplicar la función switch"
            return bug_handler.default_on_error(flow_id, node_key, msg)

        try:
            conditions = []
            script.append("conditions = []")
            outputs = []
            script.append("outputs = []")
            for case in cases:
                # Se valida el output
                if "output" not in case or case["output"] == "":
                    raise Exception("Para algún case falta la salida por defecto")

                query = ""
                output = case["output"]
                for condition in case["conditions"]:
                    rule = f" {condition['rule']} " if "rule" in condition else ""
                    operator = condition["operator"]
                    field = condition["field"]
                    value = condition["value"]

                    # Para los que requieren función
                    if operator in self.functionApply:
                        value = (
                            f"pd.isnull(df['{field}'])"
                            if operator == "is null"
                            else f"pd.notnull(df['{field}'])"
                        )
                        query += f"{rule}{value}"
                    elif operator in self.noValueRequired:
                        value = (
                            f'df["{field}"] == ""'
                            if operator == "is empty"
                            else f'df["{field}"] != ""'
                        )
                        query += f"{rule}{value}"
                    else:
                        # Se formatean los tipos de datos
                        value = (
                            " '{}'".format(value)
                            if pd.api.types.is_string_dtype(df[field])
                            else value
                        )
                        value = (
                            " '{}'".format(value)
                            if pd.api.types.is_datetime64_any_dtype(df[field])
                            else value
                        )
                        query += f"{rule}df['{field}'] {operator}{value}"

                outputs.append(output)
                script.append(f'outputs.append("{output}")')
                conditions.append(pd.eval(query))
                script.append(f'conditions.append(pd.eval("{query}"))')

            df[new_column] = np.select(conditions, outputs, default=default_value)
            script.append(
                f'df["{new_column}"] = np.select(conditions, outputs, default="{default_value}")'
            )

        except Exception as e:
            msg = "(switch) Exception: " + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import numpy as np
import pandas as pd
import urllib
import json

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


# Database
class Database:
    def exec(self, flow_id, node_key, pin, settings):
        from sqlalchemy import create_engine, text
        import snowflake.connector
        from google.cloud import bigquery
        from google.oauth2 import service_account
        from pymongo import MongoClient
        import pyodbc

        script = []
        script.append("\n# DATABASE")

        host = settings["host"] if ("host" in settings and settings["host"] is not None) else None
        user = settings["user"] if ("user" in settings and settings["user"] is not None) else None
        password = settings["password"] if ("password" in settings and settings["password"] is not None) else ''
        port = str(settings["port"]) if ("port" in settings and settings["port"] is not None) else ''
        table = settings["table"] if ("table" in settings and settings["table"] is not None) else None
        database = settings["database"] if ("database" in settings and settings["database"] is not None) else None
        project = settings["project"] if ("project" in settings and settings["project"] is not None) else None
        query = settings["query"] if ("query" in settings and settings["query"] is not None) else None
        source = settings["source"] if ("source" in settings and settings["source"] is not None) else None

        if not source:
            msg = "(database) No existe el tipo de Base de Datos"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        try:
            script.append("\n# DATABASE")

            # Debe ser uno de esos recursos, debe venir host, user y pass, y debe venir tabla y database o query
            if source == 'postgresql' or source == 'mysql' or source == 'sqlite' or source == 'oracle' or source == 'mariadb':
                password = ':'+settings["password"] if ("password" in settings and settings["password"] is not None) else ''
                port = ':'+str(settings["port"]) if ("port" in settings and settings["port"] is not None) else ''

                if not (host and database and (query or table)):
                    msg = "(database) Existen campos vacíos"
                    return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
                
                connection = self.get_url_connection(source, user, password, host, port, database)
                engine = create_engine(connection)
                df = pd.read_sql(text(query if query else f'SELECT * FROM {table}'), con=engine.connect())
                engine.dispose()

            elif source == 'sqlserver_2000':
                if not (host and user and database and (query or table)):
                    msg = "(database) Existen campos vacíos"
                    return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
                
                # ['SQL Server', 'SQL Server Native Client 11.0', 'ODBC Driver 17 for SQL Server']
                connection = pyodbc.connect("DRIVER={SQL Server};"+ f"User={user};Password={password};Database={database};Server={host};Port={port};")
                cursor = connection.cursor()
                cursor.execute(query if query else f'SELECT * FROM {table}')
                results = np.array(cursor.fetchall())
                column_names = [str(column[0]) for column in cursor.description]
                df = pd.DataFrame(results, columns=column_names)
                cursor.close()
                connection.close()

            elif source == "bigquery":
                if not (host and database and table and project):
                    msg = "(database) Existen campos vacíos"
                    return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
                
                with open(host) as file:
                    host = json.load(file)

                credentials = service_account.Credentials.from_service_account_info(host)
                client = bigquery.Client(credentials=credentials)
                table_ref = client.dataset(database, project=project).table(table)
                rows = client.list_rows(table_ref)
                df = rows.to_dataframe()
                client.close()

            elif source == "snowflake":
                if not (user and password and host and project and database):
                    msg = "(database) Existen campos vacíos"
                    return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
                
                connection = snowflake.connector.connect(user=user, password=password, account=host, database=project, schema=database)
                query = f'SELECT * FROM "{table}"'
                cursor = connection.cursor()
                cursor.execute(query)
                results = cursor.fetchall()
                column_names = [desc[0] for desc in cursor.description]
                df = pd.DataFrame(results, columns=column_names)
                connection.close()
                cursor.close()

            elif source == "mongodb":
                if not (host and database and table):
                    msg = "(database) Existen campos vacíos"
                    return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
                
                client = MongoClient(host)
                db = client[database]
                collection = db[table]
                data = list(collection.find())
                df = pd.DataFrame(data)
                client.close()

            else:
                msg = "(database) El recurso de conexión, no coincide con ninguno"
                return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        except Exception as e:
            msg = "(database) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}
    
    def get_url_connection(self, source, user, password, host, port, database):
        # if source == 'sqlserver_2000':
        #     quoted = urllib.parse.quote_plus('DRIVER={SQL Server};'+f'User={user};Password={password};Server={host};Port={port};Database={database}')    
        #     return 'mssql+pyodbc:///?odbc_connect={}'.format(quoted)
        if source == 'sqlite':
            # return f"sqlite:///:memory:"
            return f"sqlite:///{host}"
        elif source == 'mysql':
            return f"mysql+pymysql://{user}{password}@{host}{port}/{database}"
        elif source == 'postgresql':
            return f"postgresql://{user}{password}@{host}{port}/{database}"


import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Concat:
    def exec(self, flow_id, node_key, pin, settings):
        script = []

        # if node_key in cache_handler.cache[flow_id] and cache_handler.cache[flow_id][node_key]['config'] == json.dumps(settings, sort_keys=True):
        # 	bug_handler.console(f'Nodo "{node_key}" leído desde cache flow_id: "{flow_id}"', 'info', flow_id)
        # 	reset_childs = False
        # 	script_handler.script += cache_handler.cache[flow_id][node_key]['script']
        # 	return cache_handler.cache[flow_id][node_key]['pout'], reset_childs

        script.append("\n# CONCAT")
        df_A: pd.DataFrame = pin["A"].copy()
        df_B: pd.DataFrame = pin["B"].copy()
        selected_A: list = settings["a"] if "a" in settings and settings["a"] is not None else None
        selected_B: list = settings["b"] if "b" in settings and settings["b"] is not None else None

        # print(df_A.columns)
        # print(df_B.columns)

        if not selected_A:
            msg = "(concat): Debes mantener al menos un campo en el puerto A"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        df_A = df_A[selected_A]
        script.append("df_A = df_A[{}]".format(selected_A))

        # Me quedo con los campos seleccionados del Source
        if not selected_B:
            msg = "(concat): Debes mantener al menos un campo en el puerto B"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        df_B = df_B[selected_B]
        script.append("df_B = df_B[{}]".format(selected_B))

        try:
            df = pd.concat([df_A, df_B], ignore_index=True)
            script.append("df = pd.concat([df_A, df_B], ignore_index=True)")
        except Exception as e:
            msg = "(concat) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}
